package com.streambase.contrib.pricing;

import java.util.ArrayList;
import java.util.List;

import com.streambase.sb.CompleteDataType;
import com.streambase.sb.DataType;
import com.streambase.sb.client.CustomFunctionResolver;

public class Payoffs {
	
	private static final java.util.Random RAND = new java.util.Random();	
			
	/**
	 * 
	 * @param S0 initial stock price
	 * @param r interest rate
	 * @param T time to maturity
	 * @param s standard deviation of the security
	 * @param K strike price
	 * @param Trials number of simulations performed
	 * @param periods the number of exercise periods
	 * @return a list containing the terminal prices of the stock (for each trial)
	 */		
	@CustomFunctionResolver("simulatePricesFunctionResolver")
	public static ArrayList<ArrayList<Double>> simulatePrices(double S0, double r, double T, double s, double K, int Trials, int periods) {
		ArrayList<Double> periodPrices = null;
		ArrayList<Double> previousPrices = null;
		ArrayList<ArrayList<Double>> prices = new ArrayList<ArrayList<Double>>(); 
		double deltaT = T/periods, drift = (r - 0.5*s*s)*deltaT, vol = s*Math.sqrt(deltaT);
		for (int j =0; j < periods; j++){
			periodPrices = new ArrayList<Double>();
			for (int i = 0; i < Trials; i++){
				if (j == 0)
					periodPrices.add(S0*Math.exp(drift + vol*RAND.nextGaussian()));
				else
					periodPrices.add(previousPrices.get(i)*Math.exp(drift + vol*RAND.nextGaussian()));
			}
			prices.add(periodPrices);
			previousPrices = periodPrices;			
		}
		return prices;
	}
	
	/**
	 * 
	 * @param matrix a matrix of values
	 * @return the last vector in the matrix
	 */
	@CustomFunctionResolver("lastFunctionResolver")
	public static int last(List<List<Double>> matrix){
		return matrix.size() - 1;
	}
	
	/**
	 * 
	 * @param S0 initial stock price
	 * @param r interest rate
	 * @param T time to maturity
	 * @param s standard deviation of the security
	 * @param K strike price
	 * @param Trials number of simulations performed 
	 * @return a list containing the terminal prices of the stock (for each trial)
	 */
	@CustomFunctionResolver("pricesFunctionResolver")
	public static List<Double> terminalPrices(double S0, double r, double T, double s, double K, int Trials) {
		List<Double> prices = new ArrayList<Double>();
		double drift = (r - 0.5*s*s)*T, vol = s*Math.sqrt(T);
		for (int i = 0; i < Trials; i++){
			prices.add(S0*Math.exp(drift + vol*RAND.nextGaussian()));
		}
		return prices;
	}
	
	/**
	 * 
	 * @param K strike price
	 * @param prices the stock prices 
	 * @return a list containing the terminal payoffs of each of the stocks
	 */
	@CustomFunctionResolver("payoffsFunctionResolver")
	public static List<Double> putPayoffs(List<Double> prices, double K) {
		List<Double> payoffs = new ArrayList<Double>();
		for (double price : prices){
			payoffs.add(Math.max(K - price, 0));
		}
		return payoffs;
	}
	

	/**
	 * 
	 * @param K strike price
	 * @param prices the stock prices 
	 * @return a list containing the terminal payoffs of each of the stocks
	 */
	@CustomFunctionResolver("payoffsFunctionResolver")
	public static List<Double> callPayoffs(List<Double> prices, double K) {
		List<Double> payoffs = new ArrayList<Double>();
		for (double price : prices){
			payoffs.add(Math.max(price - K, 0));
		}
		return payoffs;
	}
	
	/****************************************************************************************/
	/********************************* Function Resolvers ***********************************/
	/****************************************************************************************/	
	public static CompleteDataType pricesFunctionResolver(CompleteDataType arg0,
			CompleteDataType arg1,CompleteDataType arg2,CompleteDataType arg3,CompleteDataType arg4,
			CompleteDataType arg5){
		if (arg0.getDataType() != DataType.DOUBLE)
			return null;
		if (arg1.getDataType() != DataType.DOUBLE)
			return null;
		if (arg2.getDataType() != DataType.DOUBLE)
			return null;
		if (arg3.getDataType() != DataType.DOUBLE)
			return null;
		if (arg4.getDataType() != DataType.DOUBLE)
			return null;
		if (arg5.getDataType() != DataType.INT)
			return null;
		return CompleteDataType.forList(CompleteDataType.forDouble());
	}

	public static CompleteDataType payoffsFunctionResolver(CompleteDataType arg0, CompleteDataType arg1){
		if (arg0.getElementType() != CompleteDataType.forDouble())
			return null;
		if (arg1.getDataType() != DataType.DOUBLE)
			return null;
		return CompleteDataType.forList(CompleteDataType.forDouble());
	}
	
	
	public static CompleteDataType lastFunctionResolver(CompleteDataType arg0){
		if (arg0.getElementType().getElementType() != CompleteDataType.forDouble())
			return null;
		return CompleteDataType.forInt();
	}
	
	public static CompleteDataType simulatePricesFunctionResolver(CompleteDataType arg0,
			CompleteDataType arg1,CompleteDataType arg2,CompleteDataType arg3,CompleteDataType arg4,
			CompleteDataType arg5, CompleteDataType arg6){
		if (arg0.getDataType() != DataType.DOUBLE)
			return null;
		if (arg1.getDataType() != DataType.DOUBLE)
			return null;
		if (arg2.getDataType() != DataType.DOUBLE)
			return null;
		if (arg3.getDataType() != DataType.DOUBLE)
			return null;
		if (arg4.getDataType() != DataType.DOUBLE)
			return null;
		if (arg5.getDataType() != DataType.INT)
			return null;
		if (arg6.getDataType() != DataType.INT)
			return null;
		return CompleteDataType.forList(CompleteDataType.forList(CompleteDataType.forDouble()));
	}
}
