package com.streambase.contrib.pricing;

import java.util.ArrayList;
import java.util.List;

import JSci.maths.LinearMath;
import JSci.maths.matrices.AbstractDoubleSquareMatrix;
import JSci.maths.matrices.DoubleSquareMatrix;
import JSci.maths.vectors.AbstractDoubleVector;
import JSci.maths.vectors.DoubleVector;

import com.streambase.sb.CompleteDataType;
import com.streambase.sb.DataType;
import com.streambase.sb.client.CustomFunctionResolver;


public class Regressions {
		
	/**
	 * @param X evaluate function at X
	 * @return the value of the zeroth Laguerre polynomial at X
	 */
	public static double Laguerre0(double X) {
		return Math.exp(-0.5*X);
	}

	/**
	 * @param X evaluate function at X
	 * @return the value of the first Laguerre polynomial at X
	 */
	public static double Laguerre1(double X) {
		return Math.exp(-0.5*X)*(1-X);
	}
	
	/**
	 * @param X evaluate function at X
	 * @return the value of the second Laguerre polynomial at X
	 */	
	public static double Laguerre2(double X) {
		return Math.exp(-0.5*X)*(1-2*X + 0.5*X*X);
	}

	/**
	 * 
	 * @param row the number of rows in the matrix
	 * @param col the number of columns in the matrix
	 * @return a matrix with all of its elements set to zero
	 */
	public static DoubleSquareMatrix initMatrix(int rows){
		DoubleSquareMatrix A = new DoubleSquareMatrix(rows);
		for (int i = 0; i < rows; i++){
			for (int j = 0; j < rows; j++){
				A.setElement(i,j, 0.0);
			}
		}
		return A;
	}
	
	/**
	 * 
	 * @param elements the number of elements in the vector
	 * @return a vector with all of its elements set to zero
	 */
	public static DoubleVector initVector(int elements){
		DoubleVector b= new DoubleVector(elements);
		for (int i = 0; i < elements; i++){
			b.setComponent(i, 0.0);
		}
		return b;
	}
	
	/**
	 * 
	 * @param payoffs option payoffs at a specific time period
	 * @return the continuation values at a specific time period
	 */
	@CustomFunctionResolver("continuationFunctionResolver")
	public static List<Double> continuation(List<Double> Prices, List<Double> Payoffs, double discount){
		if (Payoffs == null || Prices == null || Prices.size() != Payoffs.size())
			return null;
		AbstractDoubleSquareMatrix A = initMatrix(3);
		AbstractDoubleVector b = initVector(3);
		double cont0, cont1, cont2, inTheMoney = 0;
		double payoff, price;
		for (int i = 0; i < Prices.size(); i++){
			payoff = Payoffs.get(i);
			price = Prices.get(i);
			if (payoff > 0 ){				
				inTheMoney++;
				// Calculate the Laguerre polynomials for a payoff
				cont0 = Laguerre0(price); cont1 = Laguerre1(price); cont2 = Laguerre2(price);
				
				// Update the Laguerre matrix
				A.setElement(0, 0, A.getElement(0, 0) + cont0*cont0);
				A.setElement(0, 1, A.getElement(0, 1) + cont0*cont1);
				A.setElement(0, 2, A.getElement(0, 2) + cont0*cont2);
				
				A.setElement(1, 0, A.getElement(1, 0) + cont1*cont0);
				A.setElement(1, 1, A.getElement(1, 1) + cont1*cont1);
				A.setElement(1, 2, A.getElement(1, 2) + cont1*cont2);
				
				A.setElement(2, 0, A.getElement(2, 0) + cont2*cont0);
				A.setElement(2, 1, A.getElement(1, 1) + cont2*cont1);
				A.setElement(2, 2, A.getElement(2, 2) + cont2*cont2);		
				
				// Update the Laguerre vector
				b.setComponent(0, b.getComponent(0) + discount*cont0*payoff);
				b.setComponent(1, b.getComponent(1) + discount*cont1*payoff);
				b.setComponent(2, b.getComponent(2) + discount*cont2*payoff);
			}
		}
		// Update the Laguerre matrix and vector
		A.scalarMultiply(1/inTheMoney);
		b.scalarMultiply(1/inTheMoney);
		
		// Solve the linear system Ax=b
		AbstractDoubleVector beta = LinearMath.solve(A, b);
		double b0 = beta.getComponent(0);
		double b1 = beta.getComponent(1);
		double b2 = beta.getComponent(2);
	
		// Obtain the vector of continuation values
		List<Double> contValues = new ArrayList<Double>();		
		for (int i = 0; i < Prices.size(); i++){
			price = Prices.get(i);
			contValues.add(b0*Laguerre0(price) + b1*Laguerre1(price) + b2*Laguerre2(price));
		}
		return contValues;
	}
	
	/**
	 * 
	 * @param Payoffs the payoff of the options at a specific time
	 * @param ContinuationValues the continuation value of the option at a specific time 
	 * @return the value function which represents the higher of the two inputted values
	 */
	@CustomFunctionResolver("valueFunctionResolver")
	public static List<Double> value(List<Double> Payoffs, List<Double> ContinuationValues){
		if (Payoffs == null || ContinuationValues == null || Payoffs.size() != ContinuationValues.size())
			return null;
		List<Double> values = new ArrayList<Double>();
		for (int i = 0; i < Payoffs.size(); i++){
			values.add(Math.max(Payoffs.get(i), ContinuationValues.get(i)));
		}
		return values; 		
	}
	
	
	/****************************************************************************************/
	/********************************* Function Resolvers ***********************************/
	/****************************************************************************************/
	
	public static CompleteDataType continuationFunctionResolver(CompleteDataType arg0, CompleteDataType arg1, CompleteDataType arg2){
		if (arg0.getElementType() != CompleteDataType.forDouble())
			return null;
		if (arg1.getElementType() != CompleteDataType.forDouble())
			return null;
		if (arg2.getDataType() != DataType.DOUBLE)
			return null;		
		return CompleteDataType.forList(CompleteDataType.forDouble());
	}
	
	public static CompleteDataType valueFunctionResolver(CompleteDataType arg0, CompleteDataType arg1){
		if (arg0.getElementType() != CompleteDataType.forDouble())
			return null;
		if (arg1.getElementType() != CompleteDataType.forDouble())
			return null;
		return CompleteDataType.forList(CompleteDataType.forDouble());
	}	

}
